#!/usr/bin/env python3
import argparse
from .config import read_config
from .plugins import load_plugins

from . import __version__


def main() -> None:
    """
    Main method.
    """
    parser = argument_parse()
    args = parser.parse_args()
    config = read_config(args.config)
    input_plugins = load_plugins(config, 'Importers')
    output_plugins = load_plugins(config, 'Exporters')

    input_interfaces = initialize_interfaces(config, input_plugins)
    output_interfaces = initialize_interfaces(config, output_plugins)

    nodes = import_resources(input_interfaces)
    export_resources(output_interfaces, nodes)


def initialize_interfaces(config: dict, plugins: dict) -> dict:
    """
    Method to initialize the interfaces with the configuration.

    :param config: The configuration file content.
    :type config: dict
    :param plugins: The list of loaded plugins.
    :type plugins: dict
    :returns: The list of loaded plugins initialized.
    :rtype: dict
    """
    interfaces = []
    for interface in plugins.values():
        interfaces.append(interface(config))
    return interfaces


def import_resources(interfaces: list) -> dict:
    """
    Method to get all resources from the *input* interfaces.

    :param interfaces: The list of initialized input interfaces.
    :type interfaces: list
    :returns: The resources returned by the input interfaces.
    :rtype: dict
    """
    resources = {}
    for interface in interfaces:
        resources.update(interface.import_resources())
    return resources


def export_resources(interfaces: list, resources: dict) -> None:
    """
    Method to export the resources using the *output* interfaces.

    :param interfaces: The list of initialized output interfaces.
    :type interfaces: list
    :param resources: The resources provided by the input interfaces.
    :type resources: dict
    """
    for interface in interfaces:
        interface.export_resources(resources)


def argument_parse() -> argparse.ArgumentParser:
    """
    Method to extract the arguments from the command line.

    :returns: The argument parser.
    :rtype: argparse.ArgumentParser
    """
    parser = argparse.ArgumentParser(
        description="Generates rundeck resources "
                    "file from different API sources")

    parser.add_argument(
        'config', type=str,
        help='Configuration file')
    parser.add_argument(
        '-V', '--version', action='version',
        version='%(prog)s {}'.format(__version__),
        help='Prints version')
    return parser



import yaml
from .interfaces import ResourcesExporter
from .common import normalize_path


class YAMLExporter(ResourcesExporter):
    """
    The YAML rundeck exporter class
    """

    def __init__(self, config: dict = {}):
        self.config = config.get('YAMLExporter', None)

    def export_resources(self, dictionary: dict) -> None:
        """
        Method to save nodes' information into a rundeck
        nodes resources `YAML` file.

        :param dictionary: Rundeck formatted nodes information.
        :type dictionary: dict
        """
        yaml_file = YAMLExporter.export_path(self.config)
        if yaml_file:
            abs_path = normalize_path(yaml_file)
            with open(abs_path, 'w+') as yfile:
                yaml.dump(dictionary,
                          stream=yfile,
                          explicit_start=True,
                          default_flow_style=False)
        else:
            print(yaml.dump(dictionary,
                            explicit_start=True,
                            default_flow_style=False))

    @staticmethod
    def export_path(config: str) -> str:
        """
        Method to get the rundeck export file in absolute
        path format.

        :param config: The `Rundeck` section of the configuration.
        :type config: str
        :returns: The file path of rundeck resources to export to.
        :rtype: str
        """
        if config:
            return config.get('export_path', None)
        return None



import pkg_resources


def load_plugins(config: dict, resource_types: str) -> dict:
    """
    Method to load plugins defined in the *entry_point*
    by resource_type.

    :param config: The configuration file content.
    :type config: dict
    :param resource_types: The resource type to load (Input, Output).
    :type resource_types: str
    :returns: The plugins loaded.
    :rtype: dict
    """
    entry_points = get_plugins(resource_types)
    plugins = {}
    for section in config:
        if section in entry_points:
            plugins.update({section: entry_points[section].load()})
    return plugins


def get_plugins(resource_types: str) -> dict:
    """
    Method to get the plugins defined in the *entry_point*.

    :param resource_types: The resource type to get (Input, Output).
    :type resource_types: str
    :returns: The entry points by name.
    :rtype: dict
    """
    entry_points = {}
    for entry_point in pkg_resources.iter_entry_points(resource_types):
        entry_points.update({entry_point.name: entry_point})
    return entry_points



from chef import ChefAPI, Search
from .interfaces import ResourcesImporter
from .common import normalize_path


class ChefImporter(ResourcesImporter):
    """
    A Chef node information importer
    """

    def __init__(self, config):
        self.config = config['ChefImporter']

    def get_chef_nodes(self) -> dict:
        """

        Method to get the chef nodes information.
        :returns: The chef nodes information.
        :rtype: dict
        """
        config = ChefImporter.expand_paths(self.config)
        chef_nodes = ChefImporter.call_chef(config)
        return chef_nodes

    @staticmethod
    def call_chef(config: dict) -> dict:
        """
        Method to query the chef server.

        :param config: The chef configuration.
        :type config: dict
        :returns: The chef nodes information.
        :rtype: dict
        """
        with ChefAPI(config['url'],
                     config['user_cert_path'],
                     config['user'],
                     version=config.get('version', '0.10.8'),
                     ssl_verify=config.get('ssl_cert_path', False)):
            return Search('node', 'Node Name:*')

    @staticmethod
    def expand_paths(config: dict) -> dict:
        """
        Method to expand file paths for the user certificate
        and the chef ssl certificate.

        :param config: The chef configuration.
        :type config: dict
        :returns: The chef configuration with expanded paths.
        :rtype: dict
        """
        config['user_cert_path'] = normalize_path(config['user_cert_path'])
        config['ssl_cert_path'] = normalize_path(config['ssl_cert_path'])
        return config

    def import_resources(self) -> dict:
        """
        Method to format chef resources into rundeck resources.

        :returns: Rundeck formatted nodes resources.
        :rtype: dict
        """
        chef_nodes = self.get_chef_nodes()
        nodes = {}
        for node in chef_nodes:
            try:
                _node = {}
                _node[node['name']] = {}
                _node[node['name']]['hostname'] = node['automatic']['fqdn']
                _node[node['name']]['nodename'] = node['automatic']['hostname']
                _node[node['name']]['Environment'] = node['chef_environment']
                _node[node['name']]['osName'] = node['automatic']['platform']
                _node[node['name']]['osFamily'] = \
                    node['automatic']['platform_family']
                _node[node['name']]['osVersion'] = \
                    node['automatic']['platform_version']
                _node[node['name']]['osArch'] = \
                    node['automatic']['kernel']['machine']
                _node[node['name']]['recipes'] = \
                    ','.join(node['automatic']['recipes'])
                _node[node['name']]['roles'] = \
                    ','.join(node['automatic']['roles'])
                _node[node['name']]['tags'] = \
                    ','.join(node['automatic']['recipes'])
                _node[node['name']]['chef_tags'] = \
                    ','.join(node['normal']['tags'])
                _node[node['name']]['description'] = \
                    node['automatic']['lsb']['description']
                _username = self.config.get('rundeck_user_login', None)
                if _username:
                    _node[node['name']]['username'] = _username
            except KeyError:
                continue
            nodes.update(_node)
        return nodes



import os


def normalize_path(path: str) -> str:
    """
    Method to expand and return an absolute
    path from a normal path.

    :param path: The path to normalize.
    :type path: str
    :returns: The absolute path.
    :rtype: str
    """
    exp_path = os.path.expanduser(path)
    abs_path = os.path.abspath(exp_path)
    return abs_path


def check_file(path: str) -> str:
    """
    Method to normalize the path of a file and
    check if the file exists and is a file.
    
    :param path: The file path to check.
    :type path: str
    :returns: The absolute path of a file.
    :rtype: str
    :raises: FileNotFoundError
    """
    file_path = normalize_path(path)
    if not os.path.exists(file_path) or not os.path.isfile(file_path):
        raise FileNotFoundError
    return file_path



import configparser
from .common import check_file


def read_config(path: str) -> dict:
    """
    Method to read the init configuration file.

    :param path: The path to the init configuration file.
    :type path: str
    :returns: The content of the configuration file.
    :rtype: dict
    """
    config_path = check_file(path)
    config = configparser.ConfigParser()
    config.read(config_path)
    configuration = {}
    for section in config.sections():
        configuration[section] = {}
        for key, val in config.items(section):
            configuration[section][key] = val
    return configuration



from abc import ABCMeta, abstractmethod


class ResourcesImporter(metaclass=ABCMeta):
    """
    ResourcesImporter interface definition
    """

    @abstractmethod
    def import_resources() -> dict:
        """
        This method is expected to export the
        data into a rundeck data structure that
        can be easily saved

        :returns: Resources data.
        :rtype: dict
        """
        pass


class ResourcesExporter(metaclass=ABCMeta):
    """
    ResourcesExporter interface definition
    """

    @abstractmethod
    def export_resources():
        """
        This method is expected to save the data
        into a rundeck resources formatted file.
        """
        pass
